# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import itertools as ittl
import textwrap as text
from typing import Any
from typing import NamedTuple as named_tuple_t
from typing import Sequence

from rich.text import Text as text_t

from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.extension.type import TypeAsRichStr
from conf_ini_g.specification.base import base_t
from conf_ini_g.specification.parameter.main import parameter_t
from conf_ini_g.specification.parameter.unit import unit_t
from conf_ini_g.specification.section.unit import IsUnitSection

PARAMETERS = "parameters"


@dtcl.dataclass(init=False, repr=False, eq=False)  # Cannot have __init__ method
class controller_t(named_tuple_t):
    section: str = None
    parameter: str = None
    # All controller values are equal in role. However, this one is called primary
    # because it refers to the parameters of the section, as opposed to the parameters
    # in the alternatives.
    primary_value: Any = None

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        return (
            f"[magenta]{type(self).__name__}[/]: "
            f"[blue]{self.section}.{self.parameter}[/]={self.primary_value}"
        )


@dtcl.dataclass(repr=False, eq=False)
class section_t(base_t, list[parameter_t]):
    category: str = "Main"
    optional: bool = False
    is_growable: bool = False
    controller: controller_t = None
    alternatives: dict[str, list[parameter_t]] = None
    parameters: list[parameter_t] | None = None  # Used only during instantiation.

    def __post_init__(self) -> None:
        """"""
        if self.parameters is not None:
            self.extend(self.parameters)
            self.parameters = None

    def AddUnspecifiedParameter(self, name: str, /) -> parameter_t:
        """
        For programmatic use
        """
        if IsUnitSection(self.name):
            parameter = unit_t(
                name=name,
                definition="Unit definition",
                basic=True,
            )
        else:
            parameter = parameter_t(
                name=name,
                definition="Programmatic parameter",
                description="This parameter is not part of the specification. "
                "It was added programmatically because it was found in the INI document or "
                "passed as a command-line argument",
                basic=self.basic,
            )
        self.append(parameter)

        return parameter

    @property
    def all_parameters(self) -> Sequence[parameter_t]:
        """"""
        if self.alternatives is None:
            return self
        else:
            return tuple(ittl.chain(self, *self.alternatives.values()))

    def ActiveParameters(self, controller_value: str, /) -> Sequence[parameter_t]:
        """"""
        if controller_value == self.controller.primary_value:
            return self
        else:
            return self.alternatives[controller_value]

    @property
    def controlling_values(self) -> Sequence[Any]:
        """
        No check of controlled status; Call with care.
        """
        return (self.controller.primary_value,) + tuple(self.alternatives.keys())

    def Issues(self) -> list[str]:
        """"""
        output = super().Issues()

        valid_name_sets = [tuple(_prm.name for _prm in self)]
        if self.alternatives is not None:
            for parameters in self.alternatives.values():
                valid_name_sets.append(tuple(_prm.name for _prm in parameters))
        for valid_name_set in valid_name_sets:
            if valid_name_set.__len__() > set(valid_name_set).__len__():
                output.append(
                    f"{self.name}: Section with repeated parameter names (possibly in alternatives)"
                )

        basic = self.basic
        optional = self.optional

        if not (basic or optional):
            output.append(f"{self.name}: Section is not basic but not optional")

        if (self.controller is None) and (not optional) and (self.__len__() == 0):
            output.append(f"{self.name}: Empty mandatory section")

        n_parameters = 0
        n_basic_prms = 0
        for parameter in self.all_parameters:
            output.extend(f"[{self.name}] {_iss}" for _iss in parameter.Issues())

            n_parameters += 1
            if parameter.basic:
                n_basic_prms += 1

            if parameter.basic and not basic:
                output.append(
                    f'{parameter.name}: Basic parameter in advanced section "{self.name}"'
                )
            if optional and not parameter.optional:
                output.append(
                    f'{parameter.name}: Mandatory parameter in optional section "{self.name}"'
                )

        if (n_parameters == 0) and not self.is_growable:
            output.append(
                f"{self.name}: Section without specified parameters which does not accept unspecified parameters"
            )
        if basic and (n_parameters > 0) and (n_basic_prms == 0):
            output.append(f"{self.name}: Basic section without any basic parameters")

        control = (self.controller, self.alternatives)
        if any(_elm is None for _elm in control) and any(
            _elm is not None for _elm in control
        ):
            output.append(
                f"{self.name}: Controlled section must have both a controller and alternative parameters"
            )
        elif self.controller is not None:
            if self.controller.section == self.name:
                output.append(f"{self.name}: Section cannot be controlled by itself")
            controlling_values = self.controlling_values
            if controlling_values.__len__() > set(controlling_values).__len__():
                output.append(
                    f"{self.name}: Controlled section has duplicated controller values"
                )
            for parameter in self.all_parameters:
                if not parameter.optional:
                    output.append(
                        f'{parameter.name}: Mandatory parameter in controlled section "{self.name}"'
                    )

        return output

    def _Item(self, key: str | int, /) -> parameter_t | None:
        """"""
        if isinstance(key, int):
            return list.__getitem__(self, key)

        for parameter in self.all_parameters:
            if parameter.name == key:
                return parameter

        return None

    def __contains__(self, key: str, /) -> bool:
        """"""
        return self._Item(key) is not None

    def __getitem__(self, key: str | int, /) -> parameter_t:
        """"""
        item = self._Item(key)
        if item is None:
            raise KeyError(f"{key}: Not a parameter of section {self.name}.")

        return item

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            *text.indent(super().__rich__(), "    ").splitlines(),
            f"    [blue]Category[/]@=@{self.category}",
            f"    [blue]Optional[/]@=@{self.optional}",
            f"    [blue]Growable[/]@=@{self.is_growable}",
        ]

        if has_controller := (self.controller is not None):
            output.append(f"    [blue]Controller[/]@=@{self.controller}")

        output.extend(text.indent(_prm.__rich__(), "    ") for _prm in self)
        if has_controller:
            output.append("    With alternatives:")
            for ctl_name, parameters in self.alternatives.items():
                output.append(f"        {ctl_name}")
                for parameter in parameters:
                    output.append(text.indent(parameter.__rich__(), "            "))

        output = AlignedOnSeparator(output, "@=@", " = ")

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import PyQt6.QtCore as core
import PyQt6.QtGui as visl
import PyQt6.QtWidgets as wdgt

ALIGNED_HCENTER = core.Qt.AlignmentFlag.AlignHCenter
ALIGNED_LEFT = core.Qt.AlignmentFlag.AlignLeft
ALIGNED_RIGHT = core.Qt.AlignmentFlag.AlignRight
ALIGNED_TOP = core.Qt.AlignmentFlag.AlignTop
BASE_PALETTE = visl.QPalette.ColorRole.Base
COLOR_CYAN = visl.QColorConstants.Cyan
DIALOG_ACCEPTATION = wdgt.QDialog.DialogCode.Accepted
DIALOG_ACCEPT_OPEN = wdgt.QFileDialog.AcceptMode.AcceptOpen
DIALOG_ACCEPT_SAVE = wdgt.QFileDialog.AcceptMode.AcceptSave
DIALOG_AUTO_OVERWRITE = wdgt.QFileDialog.Option.DontConfirmOverwrite
DIALOG_MODE_ANY = wdgt.QFileDialog.FileMode.AnyFile
DIALOG_MODE_EXISTING_FILE = wdgt.QFileDialog.FileMode.ExistingFile
DIALOG_MODE_FOLDER = wdgt.QFileDialog.FileMode.Directory
FORMAT_RICH = core.Qt.TextFormat.RichText
SELECTABLE_TEXT = core.Qt.TextInteractionFlag.TextSelectableByMouse
SIZE_EXPANDING = wdgt.QSizePolicy.Policy.Expanding
SIZE_FIXED = wdgt.QSizePolicy.Policy.Fixed
SIZE_MINIMUM = wdgt.QSizePolicy.Policy.Minimum
TAB_POSITION_EAST = wdgt.QTabWidget.TabPosition.East
WIDGET_TYPE = core.Qt.WindowType.Widget



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import textwrap as text
from typing import Sequence

from rich import print as rprint
from rich.text import Text as text_t

from conf_ini_g.catalog.specification.parameter.choices import choices_t
from conf_ini_g.extension.python import SpecificationPath
from conf_ini_g.extension.type import TypeAsRichStr
from conf_ini_g.specification.parameter.type import type_t
from conf_ini_g.specification.parameter.unit import UNIT_SEPARATOR
from conf_ini_g.specification.section.main import section_t
from conf_ini_g.specification.section.unit import UNIT_SECTION


@dtcl.dataclass(init=False, repr=False, eq=False)
class config_t(list[section_t]):
    spec_path: str = None

    def __init__(self, sections: Sequence[section_t], /) -> None:
        """
        Raising exceptions is adapted here since execution cannot proceed without a
        valid specification.
        """
        list.__init__(self, sections)
        self._SetControllerChoices()  # After __init__ so that self iterator is usable
        self.spec_path = SpecificationPath(sections)

        issues = self.Issues()
        if issues.__len__() > 0:
            rprint(
                self.spec_path,
                "[red]Invalid specification[/]",
                "\n".join(issues),
                sep="\n",
            )
            raise ValueError

    def _SetControllerChoices(self) -> None:
        """"""
        for section in self:
            if (controller := section.controller) is None:
                continue

            controlling_parameter = self[controller.section][controller.parameter]
            controlling_parameter.type = type_t.NewFromType(
                choices_t.NewAnnotatedType(section.controlling_values)
            )

    def AddUnitSection(self) -> None:
        """"""
        section = section_t(
            name=UNIT_SECTION,
            definition="Unit definitions",
            description=f"Units can be used in any other section "
            f"to specify a parameter value as follows: "
            f"numerical_value{UNIT_SEPARATOR}unit, e.g., 1.5'mm.",
            basic=True,
            optional=True,
            category=UNIT_SECTION,
            is_growable=True,
        )
        self.append(section)

    @property
    def section_names(self) -> Sequence[str]:
        """"""
        return tuple(_sct.name for _sct in self)

    def Issues(self) -> list[str]:
        """"""
        output = []

        if self.__len__() == 0:
            output.append("spec_path: Empty specification")
            return output

        names = self.section_names
        if names.__len__() > set(names).__len__():
            output.append("Specification with repeated section names")

        for section in self:
            if not isinstance(section, section_t):
                output.append(
                    f"{type(section).__name__}: Invalid section type; Expected={section_t.__name__}."
                )
                continue

            output.extend(section.Issues())
            if section.controller is not None:
                if section.controller.section not in self:
                    output.append(
                        f"{section.controller.section}: "
                        f'Unspecified section declared as controller of section "{section.name}"'
                    )
                else:
                    controller_section = self[section.controller.section]
                    if controller_section.controller is not None:
                        output.append(
                            f"{section.controller.section}: "
                            f'Section controlling "{section.name}" is itself controlled'
                        )
                    if section.controller.parameter not in controller_section:
                        output.append(
                            f"{section.controller.section}.{section.controller.parameter}: "
                            f'Unspecified parameter declared as controller of section "{section.name}"'
                        )
                    else:
                        controller_parameter = controller_section[
                            section.controller.parameter
                        ]
                        if controller_parameter.optional:
                            output.append(
                                f"{section.controller.section}.{section.controller.parameter}: "
                                f'Optional parameter declared as controller of section "{section.name}"'
                            )

        return output

    def _Item(self, key: str | int, /) -> section_t | None:
        """"""
        if isinstance(key, int):
            return list.__getitem__(self, key)

        for section in self:
            if section.name == key:
                return section

        return None

    def __contains__(self, key: str, /) -> bool:
        """"""
        return self._Item(key) is not None

    def __getitem__(self, key: str | int, /) -> section_t:
        """"""
        item = self._Item(key)
        if item is None:
            raise KeyError(f"{key}: Not a section of config.")

        return item

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            f"    [blue]spec_path[/]={self.spec_path}"
            f"[yellow]:{type(self.spec_path).__name__}[/]",
        ]

        for section in self:
            output.append(text.indent(section.__rich__(), "    "))

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import re as regx
from typing import Any, Sequence

from rich.text import Text as text_t

from conf_ini_g.extension.string import AsInterpretedObject
from conf_ini_g.extension.type import (
    AnnotationsOfType,
    HintTreeFromTypeHint,
    TypeHintOfAnnotated,
    TypeTemplateFromTypeHint,
    ValueMatchesTypeHint,
    annotated_hint_t,
    any_hint_h,
    complex_hint_h,
    hint_node_t,
)
from conf_ini_g.specification.parameter.annotation import annotation_t
from conf_ini_g.specification.parameter.value import INVALID_VALUE


@dtcl.dataclass(slots=True, repr=False, eq=False)
class type_t:
    template: Any
    hint: type | hint_node_t
    annotations: Sequence[annotation_t]

    @classmethod
    def NewFromType(cls, type_: any_hint_h, /) -> type_t:
        """"""
        if type_ is None:
            raise ValueError("None: Invalid type hint.")

        if type_ is Any:
            return ANY_TYPE

        # nnt: Do not use "annotations" since it shadows __future__.annotations.
        if isinstance(type_, annotated_hint_t):
            hint = TypeHintOfAnnotated(type_)
            nnt = AnnotationsOfType(type_)
        else:
            hint = type_
            nnt = ()
        template = TypeTemplateFromTypeHint(hint)
        if isinstance(hint, complex_hint_h):
            hint = HintTreeFromTypeHint(hint)

        return cls(template=template, hint=hint, annotations=nnt)

    def Issues(self) -> list[str]:
        """"""
        output = []

        for annotation in self.annotations:
            output.extend(f"[{self}] {_iss}" for _iss in annotation.Issues())

        return output

    @property
    def base_hint(self) -> type:
        """"""
        if isinstance(self.hint, hint_node_t):
            return self.hint.type
        else:
            return self.hint

    @property
    def template_as_str(self) -> str:
        """"""
        output = (
            str(self.template)
            .replace("<class '", "")
            .replace("'>", "")
            .replace(str(Ellipsis), "...")
        )
        output = regx.sub("{\d: ", "{", output, flags=regx.ASCII)
        output = regx.sub(", \d:", " |", output, flags=regx.ASCII)

        return output

    def FirstAnnotationWithAttribute(
        self, attribute: str | Sequence[str], /
    ) -> annotation_t | None:
        """"""
        # Do not test isinstance(attribute, Sequence) since str is a sequence
        if isinstance(attribute, str):
            attributes = (attribute,)
        else:
            attributes = attribute

        for annotation in self.annotations:
            if all(hasattr(annotation, _ttr) for _ttr in attributes):
                return annotation

        return None

    def ValueIsCompliant(self, value: Any, /) -> bool:
        """"""
        return all(_nnt.ValueIsCompliant(value) for _nnt in self.annotations)

    def InterpretedValueOf(self, value: Any, /) -> tuple[Any, bool]:
        """"""
        if isinstance(value, str):
            typed_value, success = AsInterpretedObject(value, expected_type=self.hint)
        else:
            typed_value = value
            success, _ = ValueMatchesTypeHint(value, self.hint)

        if success and self.ValueIsCompliant(typed_value):
            return typed_value, True

        return INVALID_VALUE, False

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [f"[blue]{self.template}[/]"]

        for annotation in self.annotations:
            output.append(type(annotation).__name__)

        return "&".join(output)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class any_type_t(type_t):
    @classmethod
    def NewFromType(cls, type_: any_hint_h, /) -> type_t:
        """"""
        raise RuntimeError(
            f"{cls.__name__}: Not meant to be instantiated beside " f"singleton object."
        )

    def Issues(self) -> list[str]:
        """"""
        return []

    @property
    def base_hint(self) -> type:
        """"""
        return object

    def FirstAnnotationWithAttribute(
        self, attribute: str | Sequence[str], /
    ) -> annotation_t | None:
        """"""
        return None

    def ValueIsCompliant(self, value: Any, /) -> bool:
        """"""
        return True

    def InterpretedValueOf(self, value: Any, /) -> tuple[Any, bool]:
        """"""
        return value, True


ANY_TYPE = any_type_t(template="object", hint=object, annotations=())



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from types import EllipsisType, GenericAlias, NoneType, UnionType
from typing import Annotated, Any, Iterable, Sequence, get_args, get_origin

from conf_ini_g.extension.python import SpecificationPath

complex_hint_h = GenericAlias | UnionType
raw_hint_h = type | Any | complex_hint_h
annotated_hint_t = type(Annotated[object, None])
any_hint_h = raw_hint_h | annotated_hint_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class hint_node_t:
    type: type | EllipsisType | None  # None=OR
    elements: tuple[hint_node_t, ...] | None = None


@dtcl.dataclass(slots=True, repr=False, eq=False)
class value_node_t:
    consolidated: Any
    type: type | None = None
    elements: tuple[value_node_t, ...] | None = None

    def __post_init__(self) -> None:
        """"""
        self.type = type(self.consolidated)


def TypeHintOfAnnotated(annotated_hint: annotated_hint_t, /) -> raw_hint_h:
    """"""
    return annotated_hint.__args__[0]


def AnnotationsOfType(annotated_hint: annotated_hint_t, /) -> Sequence[Any]:
    """"""
    return tuple(annotated_hint.__metadata__)


def HintTreeFromTypeHint(type_hint: raw_hint_h | EllipsisType | None, /) -> hint_node_t:
    """
    Note that type hints cannot translate into hint trees with an OR-node having a child
    OR-node. For example: str | (int | float) is interpreted as str | int | float. This
    is important when creating a type selector for multitype parameters since only
    direct child nodes are taken into account for widget creation, so these nodes must
    be types, not an OR subtree.
    """
    if type_hint is None:
        return hint_node_t(type=NoneType)

    if (origin := get_origin(type_hint)) is None:
        return hint_node_t(type=type_hint)

    # Handled types: list, set, tuple, with sets using the dict delimiters { and }.
    if origin is dict:
        raise TypeError(f"{origin.__name__}: Unhandled type.")

    elements = tuple(HintTreeFromTypeHint(_elm) for _elm in get_args(type_hint))

    if origin is UnionType:
        return hint_node_t(type=None, elements=elements)

    return hint_node_t(type=origin, elements=elements)


def TypeTemplateFromTypeHint(type_hint: raw_hint_h | EllipsisType | None, /) -> type:
    """"""
    if type_hint is None:
        return NoneType

    if (origin := get_origin(type_hint)) is None:
        return type_hint

    # Handled types: list, set, tuple, with sets using the dict delimiters { and }.
    if origin is dict:
        raise TypeError(f"{origin.__name__}: Unhandled type.")

    elements = tuple(TypeTemplateFromTypeHint(_elm) for _elm in get_args(type_hint))

    if origin is UnionType:
        return {_key: _elm for _key, _elm in enumerate(elements)}

    return origin(elements)


def ValueMatchesTypeHint(
    value: Any, type_hint: raw_hint_h | hint_node_t, /
) -> tuple[bool, value_node_t]:
    """"""
    value_tree = _ValueTreeOfValue(value)

    if type_hint is Any:
        return True, value_tree

    if isinstance(type_hint, hint_node_t):
        hint_tree = type_hint
    elif isinstance(type_hint, complex_hint_h):
        hint_tree = HintTreeFromTypeHint(type_hint)
    else:
        return isinstance(value, type_hint), value_tree

    return _CastValueTree(value_tree, hint_tree), value_tree


def ValueFromValueTree(value_tree: value_node_t, /) -> Any:
    """"""
    value = value_tree.consolidated

    if value is None:
        return None

    if isinstance(value, Iterable) and not isinstance(value, str):
        elements = (ValueFromValueTree(_elm) for _elm in value_tree.elements)
        return value_tree.type(elements)

    return value_tree.type(value)


def TypeAsRichStr(instance: Any, /, *, relative_to_home: bool = True) -> str:
    """"""
    return (
        f"[bold magenta]{type(instance).__name__}[/]"
        f"[gray]@"
        f"{SpecificationPath(type(instance), relative_to_home=relative_to_home)}:[/]"
    )


def NameValueTypeAsRichStr(name: str, value: Any, /, *, separator: str = "=") -> str:
    """"""
    if isinstance(value, Sequence) and (value.__len__() == 0):
        value = "[cyan]<empty>[/]"

    return f"[blue]{name}[/]{separator}{value}[yellow]:{type(value).__name__}[/]"


def _ValueTreeOfValue(value: Any, /) -> value_node_t:
    """"""
    if isinstance(value, Iterable) and not isinstance(value, str):
        elements = tuple(_ValueTreeOfValue(_elm) for _elm in value)
        return value_node_t(consolidated=value, elements=elements)

    return value_node_t(consolidated=value)


def _CastValueTree(value_node: value_node_t, hint_node: hint_node_t, /) -> bool:
    """"""
    hn_type = hint_node.type
    hn_elements = hint_node.elements

    if hn_type is Any:
        return True

    if hn_type is NoneType:
        if value_node.consolidated is None:
            return True
        else:
            return False

    if hn_type is None:  # Or
        if any(_CastValueTree(value_node, _elm) for _elm in hn_elements):
            return True
        else:
            return False

    if not isinstance(value_node.consolidated, hn_type):
        try:
            converted = hn_type(value_node.consolidated)
            success = str(converted).replace(" ", "") == str(
                value_node.consolidated
            ).replace(" ", "")
        except:
            success = False
        if not success:
            return False
    if (value_node.elements is None) and (hn_elements is None):
        value_node.type = hn_type
        return True

    if (value_node.elements is None) or (hn_elements is None):
        return False

    n_value_children = value_node.elements.__len__()
    n_hint_elements = hn_elements.__len__()
    has_ellipsis = (n_hint_elements == 2) and (hn_elements[1].type is Ellipsis)
    should_fake_ellipsis = (n_hint_elements == 1) and issubclass(hn_type, (list, set))
    if has_ellipsis or should_fake_ellipsis or (n_value_children == n_hint_elements):
        if has_ellipsis or should_fake_ellipsis:
            hint_elements = n_value_children * (hn_elements[0],)
        else:
            hint_elements = hn_elements
        for value_elm, hint_elm in zip(value_node.elements, hint_elements):
            if not _CastValueTree(value_elm, hint_elm):
                return False

        value_node.type = hn_type
        return True

    return False



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import ast as bstr
import textwrap as text
from pathlib import Path as path_t
from typing import Any, Sequence

from conf_ini_g.extension.type import (
    ValueFromValueTree,
    ValueMatchesTypeHint,
    complex_hint_h,
    hint_node_t,
    raw_hint_h,
)


class _none_expected_t:
    pass


EXPECTS_NONE = _none_expected_t()

TRUE_VALUES = ("true", "yes", "on")
FALSE_VALUES = ("false", "no", "off")


def Flattened(string: str, /) -> str:
    """"""
    return text.dedent(string).replace("\n", "; ")


def AlignedOnSeparator(
    string: str | Sequence[str], separator: str, replacement: str, /
) -> str | tuple[str, ...] | list[str]:
    """"""
    if should_return_str := isinstance(string, str):
        lines = string.splitlines()
    else:
        lines = string
    indices = tuple(_lne.find(separator) for _lne in lines)
    longest = max(indices)

    output = (
        _lne.replace(separator, (longest - _lgt) * " " + replacement, 1)
        if _lgt > 0
        else _lne
        for _lne, _lgt in zip(lines, indices)
    )
    if should_return_str:
        return "\n".join(output)
    elif isinstance(string, tuple):
        return tuple(output)
    else:
        return list(output)


def AsInterpretedObject(
    string: str,
    /,
    *,
    expected_type: raw_hint_h | hint_node_t | None = None,
) -> tuple[Any, bool]:
    """
    expected_type: Must not be passed explicitly as None since None is interpreted as
    "no specific expected type". When expecting None, pass EXPECTS_NONE.
    """
    if expected_type is None:
        return _AsInterpretedObjectWithoutClue(string)

    if isinstance(expected_type, (complex_hint_h, hint_node_t)):
        value, _ = _AsInterpretedObjectWithoutClue(string)
        success, value_tree = ValueMatchesTypeHint(value, expected_type)
        if success:
            return ValueFromValueTree(value_tree), True
        else:
            return None, False

    return _AsInterpretedObject(string, expected_type)


def _AsInterpretedObjectWithoutClue(string: str, /) -> tuple[Any, bool]:
    """"""
    try:
        value = bstr.literal_eval(string)
    except (SyntaxError, ValueError):
        value = string

    return value, True


def _AsInterpretedObject(string: str, expected_type: type, /) -> tuple[Any, bool]:
    """"""
    failed_interpretation = None, False
    lowered = string.lower()

    if expected_type is EXPECTS_NONE:
        return None, (lowered == "none")

    if expected_type is bool:
        if lowered in TRUE_VALUES:
            return True, True
        if lowered in FALSE_VALUES:
            return False, True
        return failed_interpretation

    if expected_type is path_t:
        if string.__len__() > 0:
            return path_t(string), True
        else:
            return None, True

    # The expected type might be instantiable from a string, e.g.: float("1.0").
    # However, a success does not mean that the interpretation is valid, e.g.:
    # tuple("(1, 2, 3)"). To confirm that a success is indeed a correct interpretation,
    # the string representation of the interpreted value is compared with the string.
    # This is not a perfect test, so "literal_eval" might still be called below.
    try:
        value = expected_type(string)
        success = str(value).replace(" ", "") == string.replace(" ", "")
    except:
        value, success = failed_interpretation
    if success:
        return value, True

    try:
        value = bstr.literal_eval(string)
        success = type(value) is expected_type
        if not success:
            value = None
    except (SyntaxError, ValueError):
        value, success = failed_interpretation

    return value, success
